<?php
/**
 * @file
 * Contains migrate source plugin for XLS files.
 */

namespace Drupal\migrate_source_xls\Plugin\migrate\source;

use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate\MigrateException;
use Drupal\migrate\Plugin\migrate\source\SourcePluginBase;
use Drupal\migrate\Row;

/**
 * Source plugin for XLS files.
 *
 * @MigrateSource(
 *   id = "xls"
 * )
 */
class XlsSource extends SourcePluginBase {
  /**
   * PHPExcel file.
   *
   * @var \PHPExcel
   */
  protected $file;
  /**
   * Fields array.
   *
   * @var array
   */
  protected $fields;

  /**
   * XlsSource constructor.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, MigrationInterface $migration) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $migration);
    // Path is required.
    if (empty($this->configuration['path'])) {
      throw new MigrateException('You must declare the "path" to the source xls file in your source settings.');
    }
    // Convert URI to realpath.
    $this->configuration['path'] = self::convertPath($this->configuration['path']);

    // Key field(s) are required.
    if (empty($this->configuration['keys'])) {
      throw new MigrateException('You must declare "keys" as a unique array of fields in your source settings.');
    }

    // Header row are required.
    if (empty($this->configuration['header_row'])) {
      throw new MigrateException('You must declare "header_row" index to let source plugin know what row is a header.');
    }
    // Prepare PHPExcel file.
    $this->file = \PHPExcel_IOFactory::createReaderForFile($this->configuration['path'])
      ->load($this->configuration['path']);
    // Prepare fields array.
    $this->fields = $this->fields();
  }

  /**
   * Convert URI path to realpath.
   *
   * @param string $uri
   *   File URI.
   *
   * @return string
   *   Realpath of the source file.
   */
  private static function convertPath($uri) {
    return \Drupal::service('file_system')->realpath($uri);
  }

  /**
   * Return a string representing the source query.
   *
   * @return string
   *   The file path.
   */
  public function __toString() {
    return $this->configuration['path'];
  }

  /**
   * {@inheritdoc}
   */
  public function initializeIterator() {
    $excel = $this->file;
    if ($this->configuration['sheet_name']) {
      $iterator = $excel->setActiveSheetIndexByName($this->configuration['sheet_name']);
    }
    else {
      $iterator = $excel->setActiveSheetIndex(0);
    }
    return $iterator
      ->getRowIterator()
      ->resetStart($this->configuration['header_row']);
  }

  /**
   * {@inheritdoc}
   */
  public function next() {
    $this->currentSourceIds = NULL;
    $this->currentRow = NULL;

    // In order to find the next row we want to process, we ask the source
    // plugin for the next possible row.
    while (!isset($this->currentRow) && $this->getIterator()->valid()) {
      /** @var \PHPExcel_Worksheet_RowIterator $iterator */
      $row_data = [];
      $iterator = $this->getIterator();
      /** @var \PHPExcel_Cell $cell */
      foreach ($iterator->current()->getCellIterator() as $cell) {
        if (isset($this->fields[$cell->getColumn()])) {
          $row_data[$this->fields[$cell->getColumn()]] = $cell->getValue();
        }
      }
      $iterator->next();
      $row = new Row($row_data, $this->migration->getSourcePlugin()->getIds(), $this->migration->getDestinationIds());

      // Populate the source key for this row.
      $this->currentSourceIds = $row->getSourceIdValues();

      // Pick up the existing map row, if any, unless getNextRow() did it.
      if (!$this->mapRowAdded && ($id_map = $this->idMap->getRowBySource($this->currentSourceIds))) {
        $row->setIdMap($id_map);
      }

      // Clear any previous messages for this row before potentially adding
      // new ones.
      if (!empty($this->currentSourceIds)) {
        $this->idMap->delete($this->currentSourceIds, TRUE);
      }

      // Preparing the row gives source plugins the chance to skip.
      if ($this->prepareRow($row) === FALSE) {
        continue;
      }

      // Check whether the row needs processing.
      // 1. This row has not been imported yet.
      // 2. Explicitly set to update.
      // 3. The row is newer than the current highwater mark.
      // 4. If no such property exists then try by checking the hash of the row.
      if (!$row->getIdMap() || $row->needsUpdate() || $this->aboveHighwater($row) || $this->rowChanged($row)) {
        $this->currentRow = $row->freezeSource();
      }
    }
  }
  
  /**
   * {@inheritdoc}
   */
  public function getIDs() {
    $ids = [];
    foreach ($this->configuration['keys'] as $key) {
      $ids[$key]['type'] = 'string';
    }
    return $ids;
  }

  /**
   * {@inheritdoc}
   */
  public function count($refresh = FALSE) {
    $count = $this->file
      ->setActiveSheetIndex()
      ->getHighestRow('A');
    if ($this->configuration['offset'] > 0) {
      $count -= $this->configuration['offset'];
    }
    return $count;
  }

  /**
   * {@inheritdoc}
   */
  public function fields() {
    $fields = [];
    /** @var \PHPExcel_Worksheet_RowIterator $iterator */
    $iterator = $this->getIterator();
    /** @var \PHPExcel_Worksheet_Row $header */
    $header = $iterator->seek($this->configuration['header_row'])->current();
    /** @var \PHPExcel_Cell $cell */
    foreach ($header->getCellIterator() as $cell) {
      $value = $cell->getValue();
      // Ignore empty column headers.
      if (empty($value)) {
        continue;
      }
      // Convert column names to machine-readable format.
      $value = self::toMachineName($value);
      // Save the column names mapping.
      $fields[$cell->getColumn()] = $value;
    }
    if ($this->configuration['offset']) {
      $iterator->resetStart($this->configuration['offset'] + 1);
    }
    return $fields;
  }

  /**
   * Convert name string to machine-readable format.
   *
   * @param string $name
   *   Input name string.
   *
   * @return string
   *   Converted name string.
   */
  private static function toMachineName($name) {
    return str_replace(' ', '_', strtolower($name));
  }

}
